-------------------------------------------------------------------------------
-- Title      : Controller
-- Project    : sblock2006
-------------------------------------------------------------------------------
-- File       : controller.vhd
-- Author     : Mats Jrgen yan  <matsjorg@orakel.ntnu.no>
-- Platform   : BenBlue-III, Virtex-II Pro (XC2VP70)
-------------------------------------------------------------------------------
-- Description: Kontroller som styrer matrisen
-------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use work.sblock2006_package.ALL;

entity controller is  
	Port ( 
	 	--Kontrollsignaler
		config_matrix 			: out std_logic;
		set_matrix 				: out std_logic;
		devel 					: out std_logic;
		matrix_clear 			: out std_logic;
		store_matrix 			: out std_logic;
		run 						: out std_logic;			  
		collect_sum 			: out std_logic;

		--Data om valgt Sblock
		sblock_data_in 		: in std_logic_vector(5 downto 0);
		sblock_row 				: out std_logic_vector(7 downto 0);
		sblock_data_out 		: out std_logic_vector(5 downto 0);
		sblock_addr 			: out std_logic_vector(4 downto 0);
		sblock_write 			: out std_logic_vector(MATRIX_HEIGHT-1 downto 0);
		sblock_set_data_out 	: out set_array;

		--Regeldata
		rule_out 				: out std_logic_vector(48 downto 0);
		rule_we 					: out std_logic;
		rule_addr 				: out std_logic_vector(7 downto 0);
		used_rules_in 			: in std_logic_vector(255 downto 0);

		--Sum av matrisen
		sum 						: in std_logic_vector(11 downto 0);

		--Lut-data
		lutconfig_out 			: out std_logic_vector(31 downto 0);
		lutconfig_addr 			: out std_logic_vector(4 downto 0);
		lutconfig_we 			: out std_logic; 

		--PCI-data, routes til Com40
		pciBusy 					: in std_logic;
		pciEmpty 				: in std_logic;
		pciRW 					: out std_logic;
		pciEnable 				: out std_logic;
		pciData				 	: inout std_logic_vector(31 downto 0);

		--Reset og klokkene
		rst 						: in std_logic;
		clk 						: in std_logic;
		clk40 					: in std_logic
		);
end controller;

architecture Behavioral of controller is
	component com40
	port(
		send 				: in std_logic;
		data_send 		: in std_logic_vector(31 downto 0);
		receive 			: in std_logic;
		pciBusy 			: in std_logic;
		pciEmpty 		: in std_logic;
		rst 				: in std_logic;
		clk40 			: in std_logic;    
		pciData 			: inout std_logic_vector(31 downto 0);      
		ack_send 		: out std_logic;
		ack_receive 	: out std_logic;
		data_receive 	: out std_logic_vector(31 downto 0);
		pciRW 			: out std_logic;
		pciEnable		: out std_logic
		);
	end component;

	component instr_store
	port (
		addr				: in std_logic_VECTOR(7 downto 0);
		clk				: in std_logic;
		din				: in std_logic_VECTOR(31 downto 0);
		dout				: out std_logic_VECTOR(31 downto 0);
		we					: in std_logic
		);
	end component;
	
	--Inn-kommunikasjon med Com40
	signal data_in 				: std_logic_vector(31 downto 0);
	signal data_receive_ack 	: std_logic;
	signal data_receive			: std_logic;
	
	--Ut-kommunikasjon med Com40
	signal data_out 				: std_logic_vector(31 downto 0);
	signal data_send_ack 		: std_logic;
	signal data_send 				: std_logic;

	--Variable som brukes under kjring
	signal lastrule 				: std_logic_vector(7 downto 0); 			--Nr p siste regel
	signal wait_cycles 			: std_logic_vector(24 downto 0);			--Antall sykler til neste instruksjon
	signal instr_source 			: std_logic;									--Instruksjonskilde 1=PCI 0=BRAM
	signal instruction_store 	: std_logic;									--Bolsk verdi om instruksjonene skal lagres
	signal instr_address 		: std_logic_vector(7 downto 0);			--Adresse til instruksjon i BRAM
	signal instruction 			: std_logic_vector(31 downto 0);			--Instruksjon som skal kjres
	signal instruction_stored 	: std_logic_vector(31 downto 0); 		--Instruksjon fra BRAM
	signal devCounter 			: std_logic_vector(8 downto 0);			--Teller antall dev-steg
	signal cust_count 			: std_logic_vector(8 downto 0);			--Brukerens egen teller
	signal lastinstr_64_type 	: std_logic_vector(1 downto 0);			--Hva slags instruksjon forrige var
	signal send_mode 				: std_logic_vector(1 downto 0);			--Hva som skal sendes over PCI
	signal rule_part 				: std_logic_vector(1 downto 0);			--Brukt informasjon om hvilke 32 av 256 regler skal sendes
	
	--Synkronisering med 40MHz klokke
	signal ack_receive_i 		: std_logic;
	signal ack_receive_ii 		: std_logic;
	signal ack_send_i 			: std_logic;
	signal ack_send_ii 			: std_logic;	

begin
	--40MHz kommunikasjonsmodul
	comint: com40 port map(
		send 				=> data_send,
		ack_send			=> data_send_ack,
		data_send 		=> data_out,
		receive 			=> data_receive,
		ack_receive 	=> data_receive_ack,
		data_receive 	=> data_in,
		pciBusy 			=> pciBusy,
		pciEmpty 		=> pciEmpty,
		pciRW 			=> pciRW,
		pciEnable 		=> pciEnable,
		pciData 			=> pciData,
		rst 				=> rst,
		clk40 			=> clk40
	);
	
	--Instruksjonslager	
	instrstore : instr_store
	port map(
	addr 					=> instr_address,
	clk 					=> clk,
	din 					=> instruction,
	dout 					=> instruction_stored,
	we 					=> instruction_store
	);
   
	--Synkronisering mot 40MHz
	process (rst, clk) begin
    	if rst = '0' then
      	ack_receive_ii 	<= '0';
      	ack_receive_i 		<= '0';
    	elsif rising_edge (clk) then
      	ack_receive_ii 	<= data_receive_ack;
      	ack_receive_i 		<= ack_receive_ii;
    	end if;
  	end process;

	--Synkronisering mot 40MHz	
	process (rst, clk) begin
    	if rst = '0' then
      	ack_send_ii 	<= '0';
      	ack_send_i 		<= '0';
    	elsif rising_edge (clk) then
      	ack_send_ii 	<= data_send_ack;
      	ack_send_i 		<= ack_send_ii;
    	end if;
  	end process;

	--Utfring av intruksjoner
	process (clk, rst, ack_receive_i) begin
		--reset
		if rst = '0' then
			data_send 		<= '0';
			send_mode 		<= "00";
			data_out 		<= (others => '0');
			wait_cycles 	<= "0000000000000000000000001";
			instr_address 	<= "00000000";
			rule_we 			<= '0';
			run 				<= '0';
			lutconfig_we 	<= '0';
			matrix_clear 	<= '0';
			config_matrix 	<= '0';
			store_matrix 	<= '0';
			set_matrix 		<= '0';
			devel 			<= '0';
			collect_sum 	<= '0';
			devCounter 		<= "000000000";
			instr_source 	<= '1';
			instruction_store <= '0';
			lastinstr_64_type <= "00";
			lastrule 		<= "00000000";
		else
			if rising_edge(clk) then
				if send_mode = "00" then 			--Send Sblock-informasjon
					data_out(5 downto 0) <= sblock_data_in;
				elsif send_mode = "01" then 		--Send sum
					data_out(11 downto 0) <= sum;
				elsif send_mode = "10" then		--Send brukte regler
					data_out <= used_rules_in(CONV_INTEGER(rule_part)*32+31 downto CONV_INTEGER(rule_part)*32);
				end if;
				if instr_source = '0' then			--Les intruksjon fra BRAM
					instruction <= instruction_stored;
				elsif ack_receive_i = '1' then	--Instruksjon fra PCI
					if data_receive = '1' then		--Ny intruksjon tilgjengelig
						instr_address <= instr_address + '1';	--k lagringsadressen med 1
					end if;
					instruction <= data_in;			--Instruksjon fra PCI
					data_receive <= '0';				--Vent med  motta neste instruksjon
					wait_cycles <= "0000000000000000000000001";	--Vent en klokke med neste instruksjon
				elsif data_send = '1' then			
					if ack_send_i = '1' then		--Data er sendt, ikke send lenger
						data_send <= '0';
					end if;
					wait_cycles <= "0000000000000000000000001";
				end if;
				if instruction(5 downto 0) = "001111" then --end
					instruction_store <= '0';
				elsif instruction(5 downto 0) = "001110" then --store
					instruction_store <= '1';
					instr_address <= "00000000";
					instr_source <= '1';
				end if;
					if wait_cycles = "000000000000000000000000" then 	--Utfr en instruksjon
						if instruction_store = '0' then					 	--Ikke lagre
							if lastinstr_64_type = "01" then 				--WriteLutconv, del 2
								lutconfig_out <= instruction;
								lastinstr_64_type <= "00";	
								wait_cycles <= "0000000000000000000000001";
							elsif lastinstr_64_type = "10" then				--WriteRule, del 2
								rule_out(48 downto 17) <= instruction;
								lastinstr_64_type <= "00";
							else											  			--Ikke en 64bits instruksjon, kjr den
				  				sblock_write 	<= ZERO_PAD;
								rule_we 			<= '0';
								lutconfig_we 	<= '0';
								matrix_clear 	<= '0';
								config_matrix 	<= '0';
								set_matrix 		<= '0';
								devel 			<= '0';
								store_matrix 	<= '0';
								run 				<= '0';
								collect_sum 	<= '0';
								wait_cycles 	<= "0000000000000000000000001";
								case instruction(5 downto 0) is
								when "000000"=>				--nop
								when "001101" => 				--break
									instr_source <= '1'; 	--end, instr from pci
								when "010011" =>	 			--clearBRAM
									if instruction(28 downto 24) = "00000" then
										matrix_clear <= '1';
									end if;
								when "000111" => 				--config
									config_matrix <= '1';
									wait_cycles <= "0000000000000000000011111";
								when "001010" => 				--devel
									devel <= '1';
									wait_cycles(7 downto 0) <= lastrule+"11";
									devCounter <= devCounter + '1';
								when "101100" => 				--set
									set_matrix <= '1';
									wait_cycles <= CONV_STD_LOGIC_VECTOR(MATRIX_WIDTH,25);
								when "001100" => 				--jump
									instr_source <= '0';
									instr_address <= instruction(15 downto 8);
									wait_cycles <= "0000000000000000000000010";
								when "101001" => 				--writeLUTConv
									lutconfig_addr <= instruction(12 downto 8);
									lutconfig_we <= '1';
									lastinstr_64_type <= "01";
								when "101011" =>				--writeRule
									rule_we <= '1';
				  					rule_addr <= instruction(14 downto 7);
									rule_out(16 downto 0) <= instruction(31 downto 15);
									lastinstr_64_type <= "10";
								when "010110" => 				--jump equal
									if instruction(17) = '0' then	--devCounter
										if instruction(16 downto 8) = devCounter then 
											instr_address <= instruction(31 downto 24);
											wait_cycles <= "0000000000000000000000010";
										end if;
									else									--custCounter
										if instruction(16 downto 8) = cust_count then  
											instr_address <= instruction(31 downto 24);
											wait_cycles <= "0000000000000000000000010";
										end if;
									end if;
								when "001000" => 				--lagre matrisen
									store_matrix <= '1';
									wait_cycles <= CONV_STD_LOGIC_VECTOR(MATRIX_WIDTH-1,25);
								when "001001" => 				--run
									 run <= '1';
									 wait_cycles <= conv_std_logic_vector(conv_integer(instruction(31 downto 8)-2),25);
								when "011011" => 				--store sum
									collect_sum <= '1';
									wait_cycles <= CONV_STD_LOGIC_VECTOR(MATRIX_WIDTH*MATRIX_HEIGHT-1,25);
								when "000001" => 				--writeType
									sblock_addr <= instruction(12 downto 8);
									sblock_write(CONV_INTEGER(instruction(20 downto 16))) <= '1';
									sblock_set_data_out(CONV_INTEGER(instruction(20 downto 16)))(5) <= instruction(31);
									sblock_set_data_out(CONV_INTEGER(instruction(20 downto 16)))(4 downto 0) <= instruction(28 downto 24);
								when "010111" => 				--reset dev counter
									devCounter <= "000000000";
								when "010010" =>
									lastRule <= instruction(15 downto 8);
								when "000010" => 				--read type
									data_send <= '1';
									send_mode <= "00";
									sblock_addr <= instruction(12 downto 8);
									sblock_row <= instruction(23 downto 16);
								when "010100" => 				--read sum
									data_send <= '1';
									send_mode <= "01";
								when "010101" => 				--read used rules
									rule_part <= instruction(31 downto 30);
									data_send <= '1';
									send_mode <= "10";
								when "011100" => 				--clearCustomCount
									cust_count <= "000000000";
								when "011101" => 				--incCustomCount
									cust_count <= cust_count + '1';
								when others =>
								end case;
								instruction <="00000000000000000000000000000000";
							end if;
						else
							--bare lagre
						end if;
					elsif wait_cycles = "0000000000000000000000001" then
						--Ny instruksjon fra PCI
						if instr_source = '1' and ack_receive_i = '0' then
							data_receive <= '1';
							wait_cycles <= "0000000000000000000000000";
						--instruksjon fra minnet
						elsif instr_source = '0' then
							instr_address <= instr_address+'1';
							wait_cycles <= "0000000000000000000000000";
						end if;
					else
						wait_cycles <= wait_cycles - '1';	--Tell ned ventesykler
					end if;
			end if;
		end if;
	end process;
end Behavioral;
