-------------------------------------------------------------------------------
-- Title      : Toplevel
-- Project    : sblock2006
-------------------------------------------------------------------------------
-- File       : toplevel.vhd
-- Author     : Mats Jrgen yan  <matsjorg@orakel.ntnu.no>
-- Platform   : BenBlue-III, Virtex-II Pro (XC2VP70)
-------------------------------------------------------------------------------
-- Description: Toppniv sammenkobling av en Sblock-matrise
-------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use work.sblock2006_package.ALL;

library UNISIM;
use UNISIM.VComponents.all;

entity toplevel is
	port (	--Reset og klokker
				rst 				: in std_logic;
		 		clk 				: in std_logic;
				clk40				: in std_logic;		 

				--PCI-kommunikasjon
  		      pciBusy   		: in    std_logic;
			   pciEmpty  		: in    std_logic;
			   pciRW     		: out   std_logic;
			   pciEnable 		: out   std_logic;
			   pciData   		: inout std_logic_vector(31 downto 0)
		 );
end toplevel;

architecture struct of toplevel is
	--Definerer to typer som brukes senere for sammenkobling av Sblocker, og regler
	type std_logic_matrix_type is array (MATRIX_HEIGHT-1 downto 0) of std_logic_vector((MATRIX_WIDTH*5)-1 downto 0);
	type rule_array is array (MATRIX_HEIGHT-1 downto 0) of std_logic_vector(RULE_SIZE-1 downto 0);

	component sblock is
	port(   Ntype 				: in std_logic_vector(4 downto 0);
           Stype			   : in std_logic_vector(4 downto 0);
           Wtype 				: in std_logic_vector(4 downto 0);
           Etype 				: in std_logic_vector(4 downto 0);
           
			  Nval  				: in std_logic;
           Sval  				: in std_logic;
           Wval  				: in std_logic;
           Eval  				: in std_logic;    	
           
			  clk   				: in std_logic;
		 	  rst   				: in std_logic;

           dev_enable    	: in std_logic;
           step_enable   	: in std_logic;
		 	  config_enable 	: in std_logic;
		 
		 	  rule 				: in std_logic_vector(RULE_SIZE-1 downto 0);
		 	  
			  t2lc_data_in		: in std_logic_vector(31 downto 0);
		     t2lc_we 			: in std_logic;		 
		     t2lc_addr_in		: in std_logic_vector(4 downto 0);		 	  
		     
			  set_enable		: in std_logic;
		     set_value			: in std_logic;
		     set_type			: in std_logic_vector(TYPE_SIZE-1 downto 0);

		     rule_used			: out std_logic;
           
		 	  outval 			: out std_logic;
		 	  outtype 			: out std_logic_vector(4 downto 0)
		 );
	end component;
	component rulestore is
	port(  clk 					: in std_logic;
    		 write_enable 		: in std_logic;
          address 			: in std_logic_vector(7 downto 0);
          read_enable 		: in std_logic;
          rule_in 			: in std_logic_vector(48 downto 0);
			 rule_out 			: out std_logic_vector(48 downto 0)
	);
	end component;
	component rulecounter is
	port( clk 					: in std_logic;
		   sclr					: in std_logic;
		   q 						: out std_logic_vector(7 downto 0)
	);
	end component;
	component matrixbram is
	port( addr					: in std_logic_VECTOR(4 downto 0);
			clk					: in std_logic;
			din					: in std_logic_VECTOR(5 downto 0);
			sinit					: in std_logic;
			we						: in std_logic;
			dout					: out std_logic_VECTOR(5 downto 0)
	);
	end component;
	component sblock_counter
	port (
			q						: out std_logic_VECTOR(4 downto 0);
			ce						: in std_logic;
			clk					: in std_logic;
			sclr					: in std_logic
	);
	end component;
	component sblock_decoder
	port (
			s					: in std_logic_VECTOR(4 downto 0);
			q					: out std_logic_VECTOR(31 downto 0);
			clk				: in std_logic;
			sclr				: in std_logic;
			sset				: in std_logic
	);
	end component;
	component controller
	port(
		sblock_data_in 		: in std_logic_vector(5 downto 0);
		used_rules_in 			: in std_logic_vector(255 downto 0);
		pciBusy 					: in std_logic;
		pciEmpty 				: in std_logic;
		rst 						: in std_logic;
		clk 						: in std_logic;
		clk40 					: in std_logic;    
		pciData 					: inout std_logic_vector(31 downto 0);      
		config_matrix 			: out std_logic;
		set_matrix 				: out std_logic;
		devel 					: out std_logic;
		matrix_clear 			: out std_logic;
		store_matrix 			: out std_logic;
		run 						: out std_logic;
		collect_sum 			: out std_logic;
		sum 						: in std_logic_vector(11 downto 0);
		sblock_data_out 		: out std_logic_vector(5 downto 0);
		sblock_row 				: out std_logic_vector(7 downto 0);
		rule_out 				: out std_logic_vector(48 downto 0);
		rule_we 					: out std_logic;
		rule_addr 				: out std_logic_vector(7 downto 0);
		lutconfig_out 			: out std_logic_vector(31 downto 0);
		lutconfig_addr 		: out std_logic_vector(4 downto 0);
		lutconfig_we 			: out std_logic;
		sblock_addr 			: out std_logic_vector(4 downto 0);
		sblock_write 			: out std_logic_vector(MATRIX_HEIGHT-1 downto 0);
		sblock_set_data_out 	: out set_array;
		pciRW 					: out std_logic;
		pciEnable 				: out std_logic
		);
	end component;

	--Tilstands-signaler
	signal config 			: std_logic;
	signal set				: std_logic;
	signal devel			: std_logic;
	signal step				: std_logic;
	signal sblock_store 	: std_logic;
	signal clear_sblock 	: std_logic;

	--Et array med regler for alle kolonner
	signal rule 				: rule_array;
	
	--Regeldata fra kontrolleren
	signal com_rule_mem_write_in	: std_logic;
	signal com_rule_mem_rule 		: std_logic_vector(RULE_SIZE-1 downto 0);
	signal com_rule_address  		: std_logic_vector(7 downto 0);
	
	--Sblockdata fra kontrolleren
	signal com_sblock_in		 : std_logic_vector(5 downto 0);
	signal com_sblock_row	 : std_logic_vector(7 downto 0);
	signal com_sblock_out	 : std_logic_vector(5 downto 0);
	
	--Type-LUT-konfigdata fra kontrolleren
	signal type2lutc_data 	 : std_logic_vector(31 downto 0);
	signal type2lutc_we		 : std_logic;
	signal type2lutc_addr 	 : std_logic_vector(4 downto 0);

	--Styringssignaler for regelminnet
	signal rule_mem_address : std_logic_vector(7 downto 0);
	signal rule_read_addr 	: std_logic_vector(7 downto 0);
	signal rule_mem_read 	: std_logic;
	signal rule_mem_write	: std_logic;
	signal rule_counter_reset: std_logic;

	--Matriser med tilstandssignaler for Sblock-matrisen
	signal east_i           : std_logic_matrix;
  	signal south_i          : std_logic_matrix;
  	signal north_i          : std_logic_matrix;
  	signal west_i           : std_logic_matrix;
  	signal output_i         : std_logic_matrix;

	--Matriser med type-signaler mellom Sblocker i matrisen
	signal easttype_i			: std_logic_matrix_type;
	signal southtype_i		: std_logic_matrix_type;	
	signal northtype_i		: std_logic_matrix_type;
	signal westtype_i			: std_logic_matrix_type;
	signal outtype_i		  	: std_logic_matrix_type;
	
	--Addressering for minnet nr Sblocker skal lagres
	signal matr_set_addr				: std_logic_vector(4 downto 0);	  
	--Addressering fra kontroller
	signal matr_set_addr_cont 		: std_logic_vector(4 downto 0);  
	
	--Set-data for matrisen
	signal sblock_set_data_in		: set_array;
	--Set-data fra kontroller
	signal sblock_set_data_cont 	: set_array;
	signal sblock_set_data_out		: set_array;
	--Teller hvilken Sblock som skal lagres
	signal sblock_set_count			: std_logic_vector(4 downto 0);
	--Lagrings- og set-kontrollsignaler for matrisen
	signal sblock_write				: std_logic_vector(MATRIX_HEIGHT-1 downto 0);
	signal sblock_write_cont 		: std_logic_vector(MATRIX_HEIGHT-1 downto 0);
	signal sblock_set_en				: std_logic_vector(31 downto 0);
	signal sblock_no_set				: std_logic;
	signal sblock_counter_clr 		: std_logic;
	signal sblock_counter_en  		: std_logic;

	--Signaler for summering av sblocker
	signal collect_sum		: std_logic;
	signal sum 					: std_logic_vector(11 downto 0);
	signal sumcounter_x		: std_logic_vector(4 downto 0);
	signal sumcounter_y		: std_logic_vector(4 downto 0);

	--Signaler for lagring av brukte regler
	signal rule_used			: std_logic_matrix;
	signal rule_comp			: std_logic_matrix;
	signal rules_used			: std_logic_vector(255 downto 0);
	
begin
	--Teller hvilken regel som skal sjekkes. Starter p 0 hver dev-step
	rcount : rulecounter
		port map (
			clk				=> clk,
			sclr				=> rule_counter_reset,
			q					=> rule_read_addr
	);
	--Teller hvilken Sblock som skal konfigureres, starter p 0 hver gang set er hy
	sb_counter : sblock_counter
		port map (
			q 					=> sblock_set_count,
			ce					=> sblock_counter_en,
			clk 				=> clk,
			sclr				=> sblock_counter_clr
	);
	--Dekoder som setter riktig set-signal p Sblock-matrisen
	sb_dec : sblock_decoder
		port map (
			s 					=> sblock_set_count,
			q					=> sblock_set_en,
			clk				=> clk,
			sclr				=> sblock_no_set,
			sset				=> clear_sblock
	);

	--Generer Sblock-matrisen
	sblock_rows: for i in 0 to MATRIX_HEIGHT-1 generate
		--En BRAM for hver rad
		rs : rulestore
		port map (
			clk				=> clk,
			write_enable	=> rule_mem_write,
			address			=> rule_mem_address,
			rule_out			=> rule(i),
			read_enable		=> rule_mem_read,
			rule_in			=> com_rule_mem_rule
		);
		--En BRAM for lagring av Sblocker pr rad
		rowram : matrixbram
		port map (
			addr				=> matr_set_addr,
			clk				=> clk,
			din				=> sblock_set_data_in(i),
			sinit				=> clear_sblock,
			we					=> sblock_write(i),
			dout				=> sblock_set_data_out(i)
		);
		--Generer alle kolonnene
		sblock_cols: for j in 0 to MATRIX_WIDTH-1 generate
			--Sblock i posisjon i,j
			sblock_inst : sblock
			port map (
				Ntype				=> northtype_i(i)(j*5+4 downto j*5),
				Stype				=> southtype_i(i)(j*5+4 downto j*5),
				Wtype				=> westtype_i(i)(j*5+4 downto j*5),
				Etype				=> easttype_i(i)(j*5+4 downto j*5),
				
				Nval				=> north_i(i)(j),
				Sval				=> south_i(i)(j),
				Wval				=> west_i(i)(j),
				Eval				=> east_i(i)(j),
				
				clk				=> clk,
				rst				=> rst,
				
				dev_enable		=> devel,
				step_enable		=> step,
				config_enable	=> config,
				
				rule				=> rule(i),
				
				t2lc_data_in	=> type2lutc_data,	
		 		t2lc_we 			=> type2lutc_we,								
		 		t2lc_addr_in	=> type2lutc_addr,						

				set_enable		=> sblock_set_en(j),
				set_value		=> sblock_set_data_out(i)(5),
				set_type			=> sblock_set_data_out(i)(4 downto 0),
				
				rule_used		=> rule_used(i)(j),				

				Outval			=> output_i(i)(j),
				Outtype			=> outtype_i(i)((j*5)+4 downto (j*5))
			);
			  --reset rule_comp
			  rule_comp(i)(j) <= '0';		     

			  -- east-west routing
			  west_end: if j = 0 generate
		     		west_i(i)(j) 	<= output_i(i)(MATRIX_WIDTH-1);  -- wrap around
					westtype_i(i)(j*5+4 downto j*5) <= outtype_i(i)((MATRIX_WIDTH-1)*5+4 downto (MATRIX_WIDTH-1)*5);
		     		east_i(i)(j) 	<= output_i(i)(j+1);
					easttype_i(i)(j*5+4 downto j*5) <= outtype_i(i)((j+1)*5+4 downto (j+1)*5);
		     end generate west_end;

		     west_east: if j > 0 and j<(MATRIX_WIDTH-1) generate
		        	west_i(i)(j) 	<= output_i(i)(j-1);
					westtype_i(i)(j*5+4 downto j*5) <= outtype_i(i)((j-1)*5+4 downto (j-1)*5);
		        	east_i(i)(j) 	<= output_i(i)(j+1);
					easttype_i(i)(j*5+4 downto j*5) <= outtype_i(i)((j+1)*5+4 downto (j+1)*5);
		     end generate west_east;

		     east_end: if j = (MATRIX_WIDTH-1) generate
		        	west_i(i)(j) 	<= output_i(i)(j-1);
					westtype_i(i)(j*5+4 downto j*5) <= outtype_i(i)((j-1)*5+4 downto (j-1)*5);
		        	east_i(i)(j) 	<= output_i(i)(0);  -- wrap around
					easttype_i(i)(j*5+4 downto j*5) <= outtype_i(i)(4 downto 0);
		     end generate east_end;

		      -- north-south routing
		     north_end: if i = 0 generate
		     		north_i(i)(j) 		<= output_i(MATRIX_HEIGHT-1)(j);  -- wrap around
					northtype_i(i)(j*5+4 downto j*5) <= outtype_i(MATRIX_HEIGHT-1)(j*5+4 downto j*5);
		        	south_i(i)(j) 	<= output_i(i+1)(j);
					southtype_i(i)(j*5+4 downto j*5) <= outtype_i(i+1)(j*5+4 downto j*5);
		     end generate north_end;

		     north_south: if i > 0 and i<(MATRIX_HEIGHT-1) generate
		        	north_i(i)(j) 	<= output_i(i-1)(j);
					northtype_i(i)(j*5+4 downto j*5) <= outtype_i(i-1)(j*5+4 downto j*5);
		        	south_i(i)(j) 	<= output_i(i+1)(j);
					southtype_i(i)(j*5+4 downto j*5) <= outtype_i(i+1)(j*5+4 downto j*5);
		     end generate north_south;

		     south_end: if i = (MATRIX_HEIGHT-1) generate
		        	north_i(i)(j) 	<= output_i(i-1)(j);
					northtype_i(i)(j*5+4 downto j*5) <= outtype_i(i-1)(j*5+4 downto j*5);
		        	south_i(i)(j) 	<= output_i(0)(j);  -- wrap around
					southtype_i(i)(j*5+4 downto j*5) <= outtype_i(0)(j*5+4 downto j*5);
		     end generate south_end;

		end generate sblock_cols;
	end generate sblock_rows;	
	
	--Kontrolleren som styrer matrisen
	con: controller PORT MAP(
		config_matrix 		=> config,
		set_matrix			=> set ,
		devel 				=> devel,
		matrix_clear		=> clear_sblock,
		store_matrix 		=> sblock_store,
		run 					=> step,
		sblock_data_in 	=> com_sblock_in,
		sblock_row 			=> com_sblock_row,
		sblock_data_out 	=> com_sblock_out,
		rule_out 			=> com_rule_mem_rule,
		rule_we 				=> com_rule_mem_write_in,
		rule_addr			=> com_rule_address,
		lutconfig_out 		=> type2lutc_data,
		lutconfig_addr 	=> type2lutc_addr,
		lutconfig_we 		=> type2lutc_we,
		used_rules_in 		=> rules_used,
		rst 					=> rst,
		clk					=> clk,
		pciBusy 				=> pciBusy,
		pciEmpty 			=> pciEmpty,
		pciRW 				=> pciRW,
		pciEnable 			=> pciEnable,
		pciData 				=> pciData,
		collect_sum 		=> collect_sum,
		sblock_addr 		=> matr_set_addr_cont,
		sblock_write 		=> sblock_write_cont,
		sblock_set_data_out => sblock_set_data_cont,
		sum 					=> sum,
		clk40 				=> clk40 
	);

	--Invers set, resetter sblock-set-dekoderen
	sblock_no_set 			<= not set;
	--Enable sblock-telleren
	sblock_counter_en 	<= '1';
	--Reset regeltelleren nr det ikke er development
	rule_counter_reset 	<= not devel;
	--Send valgt Sblock til kontrolleren
	com_sblock_in 			<= sblock_set_data_out(CONV_INTEGER(com_sblock_row));
	
	--Summeringsprosess
	process(clk, collect_sum)
	begin
		if rising_edge(clk) then
			if collect_sum='1' then	--Samle inn summer
				sum <= sum + output_i(CONV_INTEGER(sumcounter_x))(CONV_INTEGER(sumcounter_y));  --Legg til valgt Sblockverdi
				if sumcounter_x = CONV_STD_LOGIC_VECTOR(MATRIX_WIDTH-1,5) then		--Sjekk om denne Sblock er siste i raden
					sumcounter_x <= "00000";													--Velg frste kolonne
					sumcounter_y <= sumcounter_y+'1';										--G til neste rad
				else											  										--Ikke enden p raden
					sumcounter_x <= sumcounter_x + '1';										--Neste kolonne
				end if;		
			else	--Ikke samle sum, resett tellerene
				sum <= "000000000000";		
				sumcounter_x <= "00000";
				sumcounter_y <= "00000";
			end if;
		end if;
	end process;

	--Lagrings- og set-prosess
	process(clk, set, sblock_store)
	begin
		if (set = '1' or sblock_store='1') then	--Lagre eller sett
			matr_set_addr <= sblock_set_count;		--Adressen til BRAM fra teller
			sblock_counter_clr<='0';					--Ikke reset telleren
		else													--Ikke lagre eller sette Sblocker
			matr_set_addr <= matr_set_addr_cont;	--Addresse fra kontroller
			sblock_counter_clr<='1';					--Resett telleren
		end if;
	end process;	

	--Regelprosess
	process(clk)
	begin
		if com_rule_mem_write_in = '1'	then			--Lagre regel
			rule_mem_write <= '1';							--Skriv
			rule_mem_address   <= com_rule_address;	--Adresse fra kontroller
		elsif devel = '1' then							  	--Kjr development
			rule_mem_write <= '0';							--Ikke skriv
			rule_mem_address   <= rule_read_addr;		--Adresse fra teller
			if rule_used = rule_comp then					--Sjekk om alle bit i rule_used = 1
			else													--En regel er brukt
				rules_used(conv_integer(rule_read_addr)-1) <= '1';	--Sett brukt regel
			end if;
 		else	--Ikke gjr development
			rules_used <= "0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000";	--Ingen brukte regler
			rule_mem_write <= '0';						--Ikke skriv
			rule_mem_address   <= rule_read_addr;	--Adresse fra teller
		end if;	
	end process;

	--Prosess for  lagre Sblocker
	process(clk,sblock_store,set)
	begin
			if sblock_store = '1' then		--Skal lagres
				sblock_write <= ONE_PAD;	--Alle Sblock-BRAM skal skrives til
				for i in MATRIX_HEIGHT-1 downto 0 loop
					sblock_set_data_in(i)(5) 					<= output_i(i)(conv_integer(sblock_set_count));	--Skriv tilstand
					sblock_set_data_in(i)(4 downto 0)	 	<= outtype_i(i)(conv_integer(sblock_set_count)*5+4 downto conv_integer(sblock_set_count)*5);	--Skriv type
				end loop;
			else
				sblock_set_data_in <= sblock_set_data_cont;	--Setdata fra kontroller
				sblock_write<=sblock_write_cont;					--Skrivesignal fra kontroller
			end if;
	end process;
end struct;
